from __future__ import absolute_import
from ...scan import Dispatcher
from ...errs import CycleDetectionError
from ...spec.v2_0.objects import Schema
from ...spec.base import NullContext
from ...spec.v2_0.parser import SchemaContext
from ...utils import deref, CycleGuard
import six

def _compose(obj, guard=None):
    guard = CycleGuard() if not guard else guard

    try:
        obj = deref(obj, guard=guard)
    except CycleDetectionError:
        return

    if obj.final != None:
        return

    if obj.items != None:
        _compose(obj.items, guard)
    for v in six.itervalues(obj.properties):
        _compose(v, guard)
    for v in (obj.allOf or []):
        _compose(v, guard)

    final = Schema(NullContext())
    final.merge(obj, SchemaContext)

    # those 'allOf' are visited by the last CycleGuard,
    # we need use a new one
    guard = CycleGuard()
    stk = list(obj.allOf)
    while len(stk) > 0:
        try:
            o = deref(stk.pop(), guard=guard)
        except CycleDetectionError:
            continue
        o = o.final if o.final else o
        final.merge(o, SchemaContext, exclude=['$ref', 'allOf'])
        for n, p in six.iteritems(o.properties):
            if n in obj.properties:
                continue
            final.properties[n] = p
        stk.extend(o.allOf or [])
    obj.update_field('final', final)


class Aggregate(object):
    """ aggregate 'allOf',
    This scanner should be run after resolving '$ref'
    """

    # TODO: rework SwaggerPrimitive.produce

    class Disp(Dispatcher): pass

    @Disp.register([Schema])
    def _schema(self, path, obj, _):
        _compose(obj)



from __future__ import absolute_import
from .consts import private
import json
import yaml
import six
import os
import logging


logger = logging.getLogger(__name__)


class Getter(six.Iterator):
    """ base of getter object

    Idealy, to subclass a getter, you just need to override load function.
    The part to extend getter would be finalized once Swagger 2.0 is ready.
    """

    def __init__(self, path):
        self.base_path = path

    def __iter__(self):
        return self

    def __next__(self):
        if len(self.urls) == 0:
            raise StopIteration

        obj = self.load(self.urls.pop(0))

        # make sure data is string type
        if isinstance(obj, six.binary_type):
            obj = obj.decode('utf-8')
        elif not isinstance(obj, six.string_types):
            raise ValueError('Unknown types: [{0}]'.format(str(type(obj))))

        # a very simple logic to distinguish json and yaml
        try:
            if obj.startswith('{'):
                obj = json.loads(obj)
            else:
                obj = yaml.load(obj)
        except ValueError:
            raise Exception('Unknown format startswith {0} ...'.format(obj[:10]))

        return obj

    def load(self, path):
        """ load the resource, and return for parsing.

        :return: name and json object of resources
        :rtype: (str, dict)
        """
        raise NotImplementedError()

class LocalGetter(Getter):
    """ default getter implmenetation for local resource file
    """
    def __init__(self, path):
        super(LocalGetter, self).__init__(path)

        for n in private.SWAGGER_FILE_NAMES:
            if self.base_path.endswith(n):
                self.base_path = os.path.dirname(self.base_path)
                self.urls = [path]
                break
            else:
                p = os.path.join(path, n)
                if os.path.isfile(p):
                    self.urls = [p]
                    break
        else:
            # there is no file matched predefined file name:
            # - resource_list.json (1.2)
            # - swagger.json       (2.0)
            # in this case, we will locate them in this way:
            # - when 'path' points to a specific file, and its
            #   extension is either 'json' or 'yaml'.
            _, ext = os.path.splitext(path)
            for e in [private.FILE_EXT_JSON, private.FILE_EXT_YAML]:
                if ext.endswith(e):
                    self.base_path = os.path.dirname(path)
                    self.urls = [path]
                    break
            else:
                for e in [private.FILE_EXT_JSON, private.FILE_EXT_YAML]:
                    if os.path.isfile(path + '.' + e):
                        self.urls = [path + '.' + e]
                        break
                else:
                    raise ValueError('Unable to locate resource file: [{0}]'.format(path))

    def load(self, path):
        ret = None

        logger.info('to load: [{0}]'.format(path))

        # try to get extension from Getter.base_path
        _, ext = os.path.splitext(self.base_path)
        # try to get extension from path
        _, ext = os.path.splitext(path) if ext == '' else (None, ext)
        # .json is default extension to try
        ext = '.json' if ext == '' else ext
        # make sure we get .json or .yaml files
        if not path.endswith(ext):
            path = path + ext

        # trim the leading slash, which is invalid on Windows
        if os.name == 'nt' and path.startswith('/'):
            path = path[1:]

        logger.info('final path to load: [{0}]'.format(path))

        with open(path, 'r') as f:
            ret = f.read()
        return ret


class UrlGetter(Getter):
    """ default getter implementation for remote resource file
    """
    def __init__(self, path):
        super(UrlGetter, self).__init__(path)
        if self.base_path.endswith('/'):
            self.base_path = self.base_path[:-1]
        self.urls = [path]

    def load(self, path):

        logger.info('to load: [{0}]'.format(path))

        ret = f = None
        try:
            f = six.moves.urllib.request.urlopen(path)
            ret = f.read()
        finally:
            if f:
                f.close()

        return ret




from __future__ import absolute_import
import six
import base64


class Byte(object):
    """ for string type, byte format
    """

    def apply_with(self, _, v, ctx):
        """ constructor

        :param str v: accept six.string_types, six.binary_type
        """
        if isinstance(v, six.binary_type):
            self.v = v
        elif isinstance(v, six.string_types):
            self.v = v.encode('utf-8')
        else:
            raise ValueError('Unsupported type for Byte: ' + str(type(v)))

    def __str__(self):
        return self.v.decode('utf-8')

    def to_json(self):
        """ according to https://github.com/wordnik/swagger-spec/issues/50,
        we should exchange 'byte' type via base64 encoding.
        """
        return base64.urlsafe_b64encode(self.v)





from __future__ import absolute_import
from ...spec.base import NullContext
from ...scan import Dispatcher
from ...errs import SchemaError
from ...utils import scope_compose, get_or_none
from ...consts import private
from ...spec.v1_2.objects import (
    ResourceList,
    Resource,
    Operation,
    Authorization,
    Parameter,
    Model,
)
from ...spec.v2_0 import objects
import os
import six


def update_type_and_ref(dst, src, scope, sep, app):
    ref = getattr(src, '$ref')
    if ref:
        dst.update_field('$ref', '#/definitions/' + scope_compose(scope, ref, sep=sep))

    if app.prim_factory.is_primitive(getattr(src, 'type', None)):
        dst.update_field('type', src.type.lower())
    elif src.type:
        dst.update_field('$ref', '#/definitions/' + scope_compose(scope, src.type, sep=sep))

def convert_min_max(dst, src):
    def _from_str(name):
        v = getattr(src, name, None)
        if v:
            if src.type == 'integer':
                # we need to handle 1.0 when converting to int
                # that's why we need to convert to float first
                dst.update_field(name, int(float(v)))
            elif src.type == 'number':
                dst.update_field(name, float(v))
            else:
                raise SchemaError('minimum/maximum is only allowed on integer/number, not {0}'.format(src.type))
        else:
            dst.update_field(name, None)

    _from_str('minimum')
    _from_str('maximum')


def convert_schema_from_datatype(obj, scope, sep, app):
    if obj == None:
        return None

    s = objects.Schema(NullContext())
    update_type_and_ref(s, obj, scope, sep, app)
    s.update_field('format', obj.format)
    if obj.is_set('defaultValue'):
        s.update_field('default', obj.defaultValue)
    convert_min_max(s, obj)
    s.update_field('uniqueItems', obj.uniqueItems)
    s.update_field('enum', obj.enum)
    if obj.items:
        i = objects.Schema(NullContext())
        update_type_and_ref(i, obj.items, scope, sep, app)
        i.update_field('format', obj.items.format)
        s.update_field('items', i)

    return s

def convert_items(o, app):
    item = objects.Items(NullContext())
    if getattr(o, '$ref'):
        raise SchemaError('Can\'t have $ref for Items')
    if not app.prim_factory.is_primitive(getattr(o, 'type', None)):
        raise SchemaError('Non primitive type is not allowed for Items')
    item.update_field('type', o.type.lower())
    item.update_field('format', o.format)

    return item


class Upgrade(object):
    """ convert 1.2 object to 2.0 object
    """
    class Disp(Dispatcher): pass

    def __init__(self, sep=private.SCOPE_SEPARATOR):
        self.__swagger = None
        self.__sep = sep

    @Disp.register([ResourceList])
    def _resource_list(self, path, obj, app):
        o = objects.Swagger(NullContext())

        #   Info Object
        info = objects.Info(NullContext())
        info.update_field('version', obj.apiVersion)
        info.update_field('title', get_or_none(obj, 'info','title'))
        info.update_field('description', get_or_none(obj, 'info', 'description'))
        info.update_field('termsOfService', get_or_none(obj, 'info', 'termsOfServiceUrl'))
        #       Contact Object
        if obj.info.contact:
            contact = objects.Contact(NullContext())
            contact.update_field('email', get_or_none(obj, 'info', 'contact'))
            info.update_field('contact', contact)
        #       License Object
        if obj.info.license or obj.info.licenseUrl:
            license = objects.License(NullContext())
            license.update_field('name', get_or_none(obj, 'info', 'license'))
            license.update_field('url', get_or_none(obj, 'info', 'licenseUrl'))
            info.update_field('license', license)

        o.update_field('info', info)

        o.update_field('swagger', '2.0')
        o.update_field('schemes', ['http', 'https'])

        o.update_field('host', '')

        o.update_field('basePath', '')
        o.update_field('tags', [])
        o.update_field('definitions', {})
        o.update_field('parameters', {})
        o.update_field('responses', {})
        o.update_field('paths', {})
        o.update_field('security', [])
        o.update_field('securityDefinitions', {})

        o.update_field('consumes', [])
        o.update_field('produces', [])

        self.__swagger = o

    @Disp.register([Resource])
    def _resource(self, path, obj, app):
        name = obj.get_name(path)
        for t in self.__swagger.tags:
            if t.name == name:
                break
        else:
            tt = objects.Tag(NullContext())
            tt.update_field('name', name)
            self.__swagger.tags.append(tt)

    @Disp.register([Operation])
    def _operation(self, path, obj, app):
        o = objects.Operation(NullContext())
        scope = obj._parent_.get_name(path)

        o.update_field('tags', [scope])
        o.update_field('operationId', obj.nickname)
        o.update_field('summary', obj.summary)
        o.update_field('description', obj.notes)
        o.update_field('deprecated', obj.deprecated == 'true')

        c = obj.consumes if obj.consumes and len(obj.consumes) > 0 else obj._parent_.consumes
        o.update_field('consumes', c if c else [])

        p = obj.produces if obj.produces and len(obj.produces) > 0 else obj._parent_.produces
        o.update_field('produces', p if p else [])

        o.update_field('parameters', [])
        o.update_field('security', [])
        # if there is not authorizations in this operation,
        # looking for it in resource object.
        _auth = obj.authorizations if obj.authorizations and len(obj.authorizations) > 0 else obj._parent_.authorizations
        if _auth:
            for name, scopes in six.iteritems(_auth):
                o.security.append({name: [v.scope for v in scopes]})

        # Operation return value
        o.update_field('responses', {})
        resp = objects.Response(NullContext())
        if obj.type != 'void':
            resp.update_field('schema', convert_schema_from_datatype(obj, scope, self.__sep, app))
        o.responses['default'] = resp

        path = obj._parent_.basePath + obj.path
        if path not in self.__swagger.paths:
            self.__swagger.paths[path] = objects.PathItem(NullContext())
 
        method = obj.method.lower()
        self.__swagger.paths[path].update_field(method, o)

    @Disp.register([Authorization])
    def _authorization(self, path, obj, app):
        o = objects.SecurityScheme(NullContext())

        if obj.type == 'basicAuth':
            o.update_field('type', 'basic')
        else:
            o.update_field('type', obj.type)
        o.update_field('scopes', {})
        for s in obj.scopes or []:
            o.scopes[s.scope] = s.description

        if o.type == 'oauth2':
            o.update_field('authorizationUrl', get_or_none(obj, 'grantTypes', 'implicit', 'loginEndpoint', 'url'))
            o.update_field('tokenUrl', get_or_none(obj, 'grantTypes', 'authorization_code', 'tokenEndpoint', 'url'))
            if o.authorizationUrl:
                o.update_field('flow', 'implicit')
            elif o.tokenUrl:
                o.update_field('flow', 'access_code')
        elif o.type == 'apiKey':
            o.update_field('name', obj.keyname)
            o.update_field('in', obj.passAs)

        self.__swagger.securityDefinitions[obj.get_name(path)] = o

    @Disp.register([Parameter])
    def _parameter(self, path, obj, app):
        o = objects.Parameter(NullContext())
        scope = obj._parent_._parent_.get_name(path)

        o.update_field('name', obj.name)
        o.update_field('required', obj.required)
        o.update_field('description', obj.description)

        if obj.paramType == 'form':
            o.update_field('in', 'formData')
        else:
            o.update_field('in', obj.paramType)

        if 'body' == getattr(o, 'in'):
            o.update_field('schema', convert_schema_from_datatype(obj, scope, self.__sep, app))
        else:
            if getattr(obj, '$ref'):
                raise SchemaError('Can\'t have $ref in non-body Parameters')

            if obj.allowMultiple == True and obj.items == None:
                o.update_field('type', 'array')
                o.update_field('collectionFormat', 'csv')
                o.update_field('uniqueItems', obj.uniqueItems)
                o.update_field('items', convert_items(obj, app))
                if obj.is_set("defaultValue"):
                    o.update_field('default', [obj.defaultValue])
                o.items.update_field('enum', obj.enum)
            else:
                o.update_field('type', obj.type.lower())
                o.update_field('format', obj.format)
                if obj.is_set("defaultValue"):
                    o.update_field('default', obj.defaultValue)
                convert_min_max(o, obj)
                o.update_field('enum', obj.enum)

            if obj.items:
                o.update_field('collectionFormat', 'csv')
                o.update_field('uniqueItems', obj.uniqueItems)
                o.update_field('items', convert_items(obj.items, app))

        path = obj._parent_._parent_.basePath + obj._parent_.path 
        method = obj._parent_.method.lower()
        op = getattr(self.__swagger.paths[path], method)
        op.parameters.append(o)

    @Disp.register([Model])
    def _model(self, path, obj, app):
        scope = obj._parent_.get_name(path)

        s = scope_compose(scope, obj.get_name(path), sep=self.__sep)
        o = self.__swagger.definitions.get(s, None)
        if not o:
            o = objects.Schema(NullContext())
            self.__swagger.definitions[s] = o

        props = {}
        for name, prop in six.iteritems(obj.properties):
            props[name] = convert_schema_from_datatype(prop, scope, self.__sep, app)
            props[name].update_field('description', prop.description)
        o.update_field('properties', props)
        o.update_field('required', obj.required)
        o.update_field('discriminator', obj.discriminator)
        o.update_field('description', obj.description)

        for t in obj.subTypes or []:
            # here we assume those child models belongs to
            # the same resource.
            sub_s = scope_compose(scope, t, sep=self.__sep)
            sub_o = self.__swagger.definitions.get(sub_s, None)
            if not sub_o:
                sub_o = objects.Schema(NullContext())
                self.__swagger.definitions[sub_s] = sub_o

            new_ref = objects.Schema(NullContext())
            new_ref.update_field('$ref', '#/definitions/' + s)
            sub_o.allOf.append(new_ref)

    @property
    def swagger(self):
        """ some preparation before returning Swagger object
        """
        # prepare Swagger.host & Swagger.basePath
        if not self.__swagger:
            return None

        common_path = os.path.commonprefix(self.__swagger.paths.keys())
        # remove tailing slash,
        # because all paths in Paths Object would prefixed with slah.
        common_path = common_path[:-1] if common_path[-1] == '/' else common_path

        if len(common_path) > 0:
            p = six.moves.urllib.parse.urlparse(common_path)
            self.__swagger.update_field('host', p.netloc)

            new_common_path = six.moves.urllib.parse.urlunparse((
                p.scheme, p.netloc, '', '', '', ''))
            new_path = {}
            for k in self.__swagger.paths.keys():
                new_path[k[len(new_common_path):]] = self.__swagger.paths[k]
            self.__swagger.update_field('paths', new_path)

        return self.__swagger



from .assign_parent import AssignParent
from .resolve import Resolve
from .patch_obj import PatchObject
from .yaml import YamlFixer
from .aggt import Aggregate
from .norm_ref import NormalizeRef
from .merge import Merge



from __future__ import absolute_import
from ..utils import deref, CycleGuard
from ._int import create_int, validate_int
from ._str import create_str, validate_str, validate_email_
from ._bool import create_bool
from ._byte import Byte
from ._time import Date, Datetime
from ._file import File
from ._float import create_float, validate_float
from ._array import Array
from ._model import Model
from ._uuid import UUID
from .comm import create_obj, _2nd_pass_obj
from .render import Renderer
import functools
import json

# TODO: enum is suitable for all types, not only string


class SwaggerPrimitive(object):
    """ primitive factory
    """
    def __init__(self):
        from ._model import Model
        from ._array import Array

        self._map = {
            # int
            'integer': {
                'int32': (create_int, validate_int),
                'int64': (create_int, validate_int),
            },

            # float
            'number':{
                'float': (create_float, validate_float),
                'double': (create_float, validate_float),
            },

            # str
            'string': {
                '': (create_str, validate_str),
                None: (create_str, validate_str),

                # TODO: add validation for email, uuid
                # TODO: add convertion of uuid from python's one
                'email': (create_str, validate_email_),
                'uuid': (functools.partial(create_obj, constructor=UUID), _2nd_pass_obj),

                'byte': (functools.partial(create_obj, constructor=Byte), _2nd_pass_obj),
                'date': (functools.partial(create_obj, constructor=Date), _2nd_pass_obj),
                'date-time': (functools.partial(create_obj, constructor=Datetime), _2nd_pass_obj),
            },

            # bool
            'boolean': {
                '': (create_bool, None),
                None: (create_bool, None),
            },

            # file
            'file': {
                '': (functools.partial(create_obj, constructor=File), _2nd_pass_obj),
                None: (functools.partial(create_obj, constructor=File), _2nd_pass_obj),
            },

            # array
            'array': {
                '': (functools.partial(create_obj, constructor=Array), _2nd_pass_obj),
                None: (functools.partial(create_obj, constructor=Array), _2nd_pass_obj),
            },

            # model / schema Object
            'object': {
                '': (functools.partial(create_obj, constructor=Model), _2nd_pass_obj),
                None: (functools.partial(create_obj, constructor=Model), _2nd_pass_obj),
            }
        }

    def get(self, _type, _format=None):
        r = self._map.get(_type, None)
        return (None, None) if r == None else r.get(_format, (None, None))

    def register(self, _type, _format, creater, _2nd_pass=None):
        """ register a type/format handler when producing primitives

        example function to create a byte primitive:
        ```python
        def create_byte(obj, val, ctx):
            # val is the value used to create this primitive, for example, a
            # dict would be used to create a Model and list would be used to
            # create an Array

            # obj in the spec used to create primitives, they are
            # Header, Items, Schema, Parameter in Swagger 2.0.

            # ctx is parsing context when producing primitives. Some primitves needs
            # multiple passes to produce(ex. Model), when we need to keep some globals
            # between passes, we should place them in ctx
            return base64.urlsafe_b64encode(val)
        ```

        example function of 2nd pass:
        ```python
        def validate_int(obj, ret, val, ctx):
            # val, obj, ctx are the same as those in creater

            # ret is the object returned by creater

            # do some stuff
            check_min_max(obj, val)

            # remember to return val, the 'outer' val would be overwritten
            # by the one you return, if you didn't return, it would be None.
            return val
        ```

        pseudo function of 2nd pass in Model:
        ```python
        def gen_mode(obj, ret, val, ctx):
            # - go through obj.properties to create properties of this model, and add
            #   them to 'ret'.
            # - remove those values used in this pass in 'val'
            # - return val
        ```

        :param _type str: type in json-schema
        :param _format str: format in json-schema
        :param creater function: a function to create a primitive.
        :param _2nd_pass function: a function used in 2nd pass when producing primitive.
        """
        if _type not in self._map:
            self._map[_type] = {}
        self._map[_type][_format] = (creater, _2nd_pass)

    def produce(self, obj, val, ctx=None):
        """ factory function to create primitives

        :param pyswagger.spec.v2_0.objects.Schema obj: spec to construct primitives
        :param val: value to construct primitives

        :return: the created primitive
        """
        val = obj.default if val == None else val
        if val == None:
            return None

        cleanup = ctx == None # it's the top-most call in recursive
        obj = deref(obj)
        ctx = {} if ctx == None else ctx
        if 'name' not in ctx and hasattr(obj, 'name'):
            ctx['name'] = obj.name
        if 'guard' not in ctx:
            ctx['guard'] = CycleGuard()
        if 'addp_schema' not in ctx:
            # Schema Object of additionalProperties
            ctx['addp_schema'] = None
        if 'addp' not in ctx:
            # additionalProperties
            ctx['addp'] = False
        if '2nd_pass' not in ctx:
            # 2nd pass processing function
            ctx['2nd_pass'] = None
        if 'factory' not in ctx:
            # primitive factory
            ctx['factory'] = self

        # cycle guard
        ctx['guard'].update(obj)

        ret = None
        if obj.type:
            creater, _2nd = self.get(_type=obj.type, _format=obj.format)
            if not creater:
                raise ValueError('Can\'t resolve type from:(' + str(obj.type) + ', ' + str(obj.format) + ')')

            ret = creater(obj, val, ctx)
            if _2nd:
                val = _2nd(obj, ret, val, ctx)
                ctx['2nd_pass'] = _2nd
        elif len(obj.properties) or obj.additionalProperties:
            ret = Model()
            val = ret.apply_with(obj, val, ctx)

        if isinstance(ret, (Date, Datetime, Byte, File)):
            # it's meanless to handle allOf for these types.
            return ret

        def _apply(o, r, v, c):
            if hasattr(ret, 'apply_with'):
                v = r.apply_with(o, v, c)
            else:
                _2nd = c['2nd_pass']
                if _2nd == None:
                    _, _2nd = self.get(_type=o.type, _format=o.format)
                if _2nd:
                    _2nd(o, r, v, c)
                    # update it back to context
                    c['2nd_pass'] = _2nd
            return v

        # handle allOf for Schema Object
        allOf = getattr(obj, 'allOf', None)
        if allOf:
            not_applied = []
            for a in allOf:
                a = deref(a)
                if not ret:
                    # try to find right type for this primitive.
                    ret = self.produce(a, val, ctx)
                    is_member = hasattr(ret, 'apply_with')
                else:
                    val = _apply(a, ret, val, ctx)

                if not ret:
                    # if we still can't determine the type,
                    # keep this Schema object for later use.
                    not_applied.append(a)
            if ret:
                for a in not_applied:
                    val = _apply(a, ret, val, ctx)

        if ret and cleanup and hasattr(ret, 'cleanup'):
            val = ret.cleanup(val, ctx)

        return ret

    def is_primitive(self, _type):
        """ check if a given object refering to a primitive
        defined in spec.

        :param dict obj: object to be checked
        :return: True if this object is a primitive.
        """
        return _type in self._map.keys()




from __future__ import absolute_import
from ...core import BaseClient
from requests import Session, Request
import six


class Client(BaseClient):
    """ Client implementation based on requests
    """

    __schemes__ = set(['http', 'https'])

    def __init__(self, auth=None):
        """
        """
        super(Client, self).__init__(auth)
        self.__s = Session()

    def request(self, req_and_resp, opt={}):
        """
        """
        req, resp = super(Client, self).request(req_and_resp, opt)

        # apply request-related options before preparation.
        req.prepare(scheme=self.prepare_schemes(req).pop(), handle_files=False)
        req._patch(opt)

        # prepare for uploaded files
        file_obj = {}
        for k, v in six.iteritems(req.files):
            f = v.data or open(v.filename, 'rb')
            if 'Content-Type' in v.header:
                file_obj[k] = (v.filename, f, v.header['Content-Type'])
            else:
                file_obj[k] = (v.filename, f)

        rq = Request(
            method=req.method.upper(),
            url=req.url,
            params=req.query,
            data=req.data,
            headers=req.header,
            files=file_obj
        )
        rq = self.__s.prepare_request(rq)
        rs = self.__s.send(rq)

        resp.apply_with(
            status=rs.status_code,
            header=rs.headers,
            raw=rs.text
        )

        return resp




from __future__ import absolute_import
from ...errs import CycleDetectionError
from ...scan import Dispatcher
from ...spec.v2_0.parser import SchemaContext
from ...spec.v2_0.objects import Schema

class Resolve(object):
    """ pre-resolve '$ref' """

    class Disp(Dispatcher): pass

    @Disp.register([Schema])
    def _schema(self, _, obj, app):
        if obj.ref_obj:
            return

        r = getattr(obj, '$ref')
        if not r:
            return

        ro = app.resolve(r, SchemaContext)
        if not ro:
            raise ReferenceError('Unable to resolve: {0}'.format(r))
        if ro.__class__ != obj.__class__:
            raise TypeError('Referenced Type mismatch: {0}'.format(r))

        obj.update_field('ref_obj', ro)





from __future__ import absolute_import
from .utils import jr_split, jp_split
from .getter import UrlGetter, LocalGetter
import six
import os
import inspect
import logging


logger = logging.getLogger(__name__)


class SwaggerResolver(object):
    """ JSON Reference Resolver:
    resolving a JSON reference to a raw object (dict),
    then return and cache it.
    """

    def __init__(self, url_load_hook):
        """
        """
        # a map from url to loaded json/yaml
        self.__cache = {}

        # things to make unittest easier,
        # all urls to load json would go through this hook
        self.__url_load_hook = url_load_hook

    def resolve(self, jref, getter=None):
        """
        """
        url, jp = jr_split(jref)

        # apply hook when use this url to load
        # note that we didn't cache SwaggerApp with this local_url
        local_url = self.__url_load_hook(url) if self.__url_load_hook else url

        logger.info('{0} patch to {1}'.format(url, local_url))

        # check cache
        obj = self.__cache.get(url, None)
        if not obj:
            # load that object
            if not getter:
                getter = UrlGetter
                p = six.moves.urllib.parse.urlparse(local_url)
                if p.scheme == 'file' and p.path:
                    getter = LocalGetter(os.path.join(p.netloc, p.path))

            if inspect.isclass(getter):
                # default initialization is passing the url
                # you can override this behavior by passing an
                # initialized getter object.
                getter = getter(local_url)

            obj = six.advance_iterator(getter)
            self.__cache[url] = obj if obj else None

        if obj:
            ts = jp_split(jp)[1:]
            while len(ts) > 0:
                t = ts.pop(0)
                if isinstance(obj, list):
                    obj = obj[int(t)]
                elif isinstance(obj, dict):
                    obj = obj[t]
                else:
                    raise Exception('Invalid type to resolve json-pointer: {0}'.format(str(type(obj))))
        else:
            raise Exception('Unable to resolve: {0}'.format(jref))

        return obj



from __future__ import absolute_import
from ..utils import jp_compose
import six
import copy
import functools
import weakref
import itertools


class ContainerType:
    """ Enum of Container-Types
    """

    # list container
    list_ = 1

    # dict container
    dict_ = 2

    # dict of list container, like: {'xx': [], 'xx': [], ...}
    dict_of_list_ = 3

def container_apply(ct, v, f, fd=None, fdl=None):
    """
    """
    ret = None
    if v == None:
        return ret

    if ct == None:
        ret = f(ct, v)
    elif ct == ContainerType.list_:
        ret = []
        for vv in v:
            ret.append(f(ct, vv))
    elif ct == ContainerType.dict_:
        ret = {}
        for k, vv in six.iteritems(v):
            ret[k] = fd(ct, vv, k) if fd else f(ct, vv)
    elif ct == ContainerType.dict_of_list_:
        ret = {}
        for k, vv in six.iteritems(v):
            if fdl:
                fdl(ct, vv, k)
            ret[k] = []
            for vvv in vv:
                ret[k].append(fd(ct, vvv, k) if fd else f(ct, vvv))
    else:
        raise ValueError('Unknown ContainerType: {0}'.format(ct))

    return ret


class Context(object):
    """ Base of all parsing contexts """

    # required fields, a list of strings
    __swagger_required__ = []

    # parsing context of children fields,
    # a list of tuple (field-name, container-type, parsing-context)
    __swagger_child__ = {}

    # factory of object to be created according to
    # this parsing context.
    __swagger_ref_obj__ = None

    def __init__(self, parent_obj, backref):
        """
        constructor

        :param dict parent_obj: parent object placeholder
        :param str backref: the key to parent object placeholder
        """

        # object placeholder of parent object
        self._parent_obj = parent_obj

        # key used to indicate the location in
        # parent object when parsing finished.
        self._backref = backref

        self.__reset_obj()

    def __enter__(self):
        return self

    def __reset_obj(self):
        self._obj = {}

    @classmethod
    def is_produced(kls, obj):
        """ Provide a hook for customizing ways to check produced output
        """
        return isinstance(obj, kls.__swagger_ref_object__)

    def produce(self):
        """ Provide a hook for customizing ways to produce output
        """
        return self.__class__.__swagger_ref_object__(self)

    def __exit__(self, exc_type, exc_value, traceback):
        """ When exiting parsing context, doing two things
        - create the object corresponding to this parsing context.
        - populate the created object to parent context.
        """
        if self._obj == None:
            return

        obj = self.produce()
        self.__reset_obj()

        if isinstance(self._parent_obj[self._backref], list):
            self._parent_obj[self._backref].append(obj)
        else:
            self._parent_obj[self._backref] = obj

    def parse(self, obj=None):
        """ major part do parsing.

        :param dict obj: json object to be parsed.
        :raises ValueError: if obj is not a dict type.
        """
        if obj == None:
            return

        if not isinstance(obj, dict):
            raise ValueError('invalid obj passed: ' + str(type(obj)))

        def _apply(x, kk, ct, v):
            if key not in self._obj:
                self._obj[kk] = {} if ct == None else []
            with x(self._obj, kk) as ctx:
                ctx.parse(obj=v)

        def _apply_dict(x, kk, ct, v, k):
            if k not in self._obj[kk]:
                self._obj[kk][k] = {} if ct == ContainerType.dict_ else []
            with x(self._obj[kk], k) as ctx:
                ctx.parse(obj=v)

        def _apply_dict_before_list(kk, ct, v, k):
            self._obj[kk][k] = []

        if hasattr(self, '__swagger_child__'):
            # to nested objects
            for key, (ct, ctx_kls) in six.iteritems(self.__swagger_child__):
                items = obj.get(key, None)

                # create an empty child, even it's None in input.
                # this makes other logic easier.
                if ct == ContainerType.list_:
                    self._obj[key] = []
                elif ct:
                    self._obj[key] = {}

                if items == None:
                    continue

                container_apply(ct, items,
                    functools.partial(_apply, ctx_kls, key),
                    functools.partial(_apply_dict, ctx_kls, key),
                    functools.partial(_apply_dict_before_list, key)
                )

        # update _obj with obj
        if self._obj != None:
            for key in (set(obj.keys()) - set(self._obj.keys())):
                self._obj[key] = obj[key]
        else:
            self._obj = obj


class BaseObj(object):
    """ Base implementation of all referencial objects,
    """

    # fields that need re-named.
    __swagger_rename__ = {}

    # dict of names of fields, we will skip fields not in this list.
    # field format:
    # - {name: default-value}: a field name with default value
    __swagger_fields__ = {}

    # fields used internally
    __internal_fields__ = {}

    # Swagger Version this object belonging to
    __swagger_version__ = None

    def __init__(self, ctx):
        """ constructor

        :param Context ctx: parsing context used to create this object
        :raises TypeError: if ctx is not a subclass of Context.
        """
        super(BaseObj, self).__init__()

        # init parent reference
        self._parent__ = None

        if not issubclass(type(ctx), Context):
            raise TypeError('should provide args[0] as Context, not: ' + ctx.__class__.__name__)

        self.__origin_keys = set([k for k in six.iterkeys(ctx._obj)])

        # handle fields
        for name, default in six.iteritems(self.__swagger_fields__):
            setattr(self, self.get_private_name(name), ctx._obj.get(name, copy.copy(default)))

        for name, default in six.iteritems(self.__internal_fields__):
            setattr(self, self.get_private_name(name), None)

        self._assign_parent(ctx)

    def _assign_parent(self, ctx):
        """ parent assignment, internal usage only
        """
        def _assign(cls, _, obj):
            if obj == None:
                return

            if cls.is_produced(obj):
                if isinstance(obj, BaseObj):
                    obj._parent__ = self
            else:
                raise ValueError('Object is not instance of {0} but {1}'.format(cls.__swagger_ref_object__.__name__, obj.__class__.__name__))

        # set self as childrent's parent
        for name, (ct, ctx) in six.iteritems(ctx.__swagger_child__):
            obj = getattr(self, name)
            if obj == None:
                continue

            container_apply(ct, obj, functools.partial(_assign, ctx))

    def get_private_name(self, f):
        """ get private protected name of an attribute

        :param str f: name of the private attribute to be accessed.
        """
        f = self.__swagger_rename__[f] if f in self.__swagger_rename__.keys() else f
        return '_' + self.__class__.__name__ + '__' + f
 
    def update_field(self, f, obj):
        """ update a field

        :param str f: name of field to be updated.
        :param obj: value of field to be updated.
        """
        n = self.get_private_name(f)
        if not hasattr(self, n):
            raise AttributeError('{0} is not in {1}'.format(n, self.__class__.__name__))

        setattr(self, n, obj)
        self.__origin_keys.add(f)

    def resolve(self, ts):
        """ resolve a list of tokens to an child object

        :param list ts: list of tokens
        """
        if isinstance(ts, six.string_types):
            ts = [ts]

        obj = self
        while len(ts) > 0:
            t = ts.pop(0)

            if issubclass(obj.__class__, BaseObj):
                obj = getattr(obj, t)
            elif isinstance(obj, list):
                obj = obj[int(t)]
            elif isinstance(obj, dict):
                obj = obj[t]

        return obj

    def merge(self, other, ctx, exclude=[]):
        """ merge properties from other object,
        - for child object, only merge from 'not-default' to 'default'.
        - for others, merge (list, dict, primitives)

        :param BaseObj other: the source object to be merged from.
        :param Context ctx: the parsing context
        """
        def _produce_new_obj(x, ct, v):
            return x(None, None).produce().merge(v, x)

        for name, default in itertools.chain(
                six.iteritems(self.__swagger_fields__),
                six.iteritems(self.__internal_fields__)):
            if name in exclude:
                continue

            v = getattr(other, name)
            if v == default:
                continue

            childs = [(n, ct, cctx) for n, (ct, cctx) in six.iteritems(ctx.__swagger_child__) if n == name]
            if len(childs) == 0:
                # we don't need to make a copy,
                # since everything under SwaggerApp should be
                # readonly.
                if isinstance(v, list):
                    try:
                        self.update_field(name, list(set((getattr(self, name) or []) + v)))
                    except TypeError:
                        self.update_field(name, list((getattr(self, name) or []) + v))
                elif isinstance(v, dict):
                    d = copy.copy(v)
                    d.update(getattr(self, name) or {})
                    self.update_field(name, d)
                elif getattr(self, name) == default:
                    self.update_field(name, v)
                continue
            else:
                # for child, stop when src object has something
                if getattr(self, name) != default:
                    continue

            ct, cctx = childs[0][1], childs[0][2]
            self.update_field(name,
                container_apply(
                    ct, v, 
                    functools.partial(_produce_new_obj, cctx)
            ))

        # make sure parent is correctly assigned.
        self._assign_parent(ctx)

        # allow cascade calling
        return self

    def is_set(self, k):
        """ check if a key is setted from Swagger API document

        :param k: the key to check
        :return: True if the key is setted. False otherwise, it means we would get value
        from default from Field.
        """
        return k in self.__origin_keys

    def compare(self, other, base=None):
        """ comparison, will return the first difference """
        if self.__class__ != other.__class__:
            return False, ''

        names = self._field_names_

        def cmp_func(name, s, o):
            # special case for string types
            if isinstance(s, six.string_types) and isinstance(o, six.string_types):
                return s == o, name

            if s.__class__ != o.__class__:
                return False, name

            if isinstance(s, BaseObj):
                if not isinstance(s, weakref.ProxyTypes):
                    return s.compare(o, name)
            elif isinstance(s, list):
                for i, v in zip(range(len(s)), s):
                    same, n = cmp_func(jp_compose(str(i), name), v, o[i])
                    if not same:
                        return same, n
            elif isinstance(s, dict):
                # compare if any key diff
                diff = list(set(s.keys()) - set(o.keys()))
                if diff:
                    return False, jp_compose(str(diff[0]), name)
                diff = list(set(o.keys()) - set(s.keys()))
                if diff:
                    return False, jp_compose(str(diff[0]), name)

                for k, v in six.iteritems(s):
                    same, n = cmp_func(jp_compose(k, name), v, o[k])
                    if not same:
                        return same, n
            else:
                return s == o, name

            return True, name

        for n in names:
            same, n = cmp_func(jp_compose(n, base), getattr(self, n), getattr(other, n))
            if not same:
                return same, n

        return True, ''

    def dump(self):
        """ dump Swagger Spec in dict(which can be
        convert to JSON)
        """
        r = {}
        def _dump_(obj):
            if isinstance(obj, dict):
                ret = {}
                for k, v in six.iteritems(obj):
                    ret[k] = _dump_(v)
                return None if ret == {} else ret
            elif isinstance(obj, list):
                ret = []
                for v in obj:
                    ret.append(_dump_(v))
                return None if ret == [] else ret
            elif isinstance(obj, BaseObj):
                return obj.dump()
            elif isinstance(obj, (six.string_types, six.integer_types)):
                return obj
            else:
                raise ValueError('Unknown object to dump: {0}'.format(obj.__class__.__name__))

        for name, default in six.iteritems(self.__swagger_fields__):
            # only dump a field when its value is not equal to default value
            v = getattr(self, name)
            if v != default:
                d = _dump_(v)
                if d != None:
                    r[name] = d

        return None if r == {} else r

    @property
    def _parent_(self):
        """ get parent object

        :return: the parent object.
        :rtype: a subclass of BaseObj.
        """
        return self._parent__

    @property
    def _field_names_(self):
        """ get list of field names defined in Swagger spec

        :return: a list of field names
        :rtype: a list of str
        """
        ret = []
        for n in six.iterkeys(self.__swagger_fields__):
            new_n = self.__swagger_rename__.get(n, None)
            ret.append(new_n) if new_n else ret.append(n)
        
        return ret

    @property
    def _children_(self):
        """ get children objects

        :rtype: a dict of children {child_name: child_object}
        """
        ret = {}
        names = self._field_names_

        def down(name, obj):
            if isinstance(obj, BaseObj):
                if not isinstance(obj, weakref.ProxyTypes):
                    ret[name] = obj
            elif isinstance(obj, list):
                for i, v in zip(range(len(obj)), obj):
                    down(jp_compose(str(i), name), v)
            elif isinstance(obj, dict):
                for k, v in six.iteritems(obj):
                    down(jp_compose(k, name), v)

        for n in names:
            down(jp_compose(n), getattr(self, n))

        return ret


def _method_(name):
    """ getter factory """
    def _getter_(self):
        return getattr(self, self.get_private_name(name))
    return _getter_


class FieldMeta(type):
    """ metaclass to init fields
    """
    def __new__(metacls, name, bases, spc):
        """ scan through MRO to get a merged list of fields,
        and create those fields.
        """
        def init_fields(fields, rename):
            for name in six.iterkeys(fields):
                name = rename[name] if name in rename.keys() else name
                spc[name] = property(_method_(name))

        def _default_(name, default):
            spc[name] = spc[name] if name in spc else default

        def _update_(dict1, dict2):
            d = {}
            for k in set(dict2.keys()) - set(dict1.keys()):
                d[k] = dict2[k]
            dict1.update(d)

        # compose fields definition from parents
        for b in bases:
            if hasattr(b, '__swagger_fields__'):
                _default_('__swagger_fields__', {})
                _update_(spc['__swagger_fields__'], b.__swagger_fields__)
            if hasattr(b, '__swagger_rename__'):
                _default_('__swagger_rename__', {})
                _update_(spc['__swagger_rename__'], b.__swagger_rename__)
            if hasattr(b, '__internal_fields__'):
                _default_('__internal_fields__', {})
                _update_(spc['__internal_fields__'], b.__internal_fields__)

        rename = spc['__swagger_rename__'] if '__swagger_rename__' in spc.keys() else {}
        # swagger fields
        if '__swagger_fields__' in spc.keys():
            init_fields(spc['__swagger_fields__'], rename)
        # internal fields
        if '__internal_fields__' in spc.keys():
            init_fields(spc['__internal_fields__'], {})

        return type.__new__(metacls, name, bases, spc)


class NullContext(Context):
    """ black magic to initialize BaseObj
    """

    _obj = None

    def __init__(self):
        super(NullContext, self).__init__(None, None)




from __future__ import absolute_import
from ..errs import ValidationError, SchemaError
import functools
import six
import json


class Array(list):
    """ for array type, or parameter when allowMultiple=True
    """

    def __init__(self):
        """ v: list or string_types
        """
        super(Array, self).__init__()
        self.__collection_format = 'csv'

    def apply_with(self, obj, val, ctx):
        """
        """
        self.__collection_format = getattr(obj, 'collectionFormat', 'csv')

        if isinstance(val, six.string_types):
            try:
                val = json.loads(val)
            except Exception as e:
                if self.__collection_format == 'csv':
                    val = val.split(',')
                elif self.__collection_format == 'ssv':
                    val = val.split(' ')
                elif self.__collection_format == 'tsv':
                    val = val.split('\t')
                elif self.__collection_format == 'pipes':
                    val = val.split('|')
                else:
                    raise e

        val = set(val) if obj.uniqueItems else val

        if obj.items and len(val):
            self.extend(map(functools.partial(ctx['factory'].produce, obj.items), val))
            val = []

        # init array as list
        if obj.minItems and len(self) < obj.minItems:
            raise ValidationError('Array should be more than {0}, not {1}'.format(obj.minItems, len(self)))
        if obj.maxItems and len(self) > obj.maxItems:
            raise ValidationError('Array should be less than {0}, not {1}'.format(obj.maxItems, len(self)))

        return val

    def __str__(self):
        """ array primitives should be for 'path', 'header', 'query'.
        Therefore, this kind of convertion is reasonable.

        :return: the converted string
        :rtype: str
        """
        def _conv(p):
            s = ''
            for v in self:
                s = ''.join([s, p if s else '', str(v)])
            return s
    
        if self.__collection_format == 'csv':
            return _conv(',')
        elif self.__collection_format == 'ssv':
            return _conv(' ')
        elif self.__collection_format == 'tsv':
            return _conv('\t')
        elif self.__collection_format == 'pipes':
            return _conv('|')
        else:
            raise SchemaError('Unsupported collection format when converting to str: {0}'.format(self.__collection_format))

    def to_url(self):
        """ special function for handling 'multi',
        refer to Swagger 2.0, Parameter Object, collectionFormat
        """
        if self.__collection_format == 'multi':
            return [str(s) for s in self]
        else:
            return [str(self)]


